import re


HTTP_GET = 'GET'
HTTP_POST = 'POST'
HTTP_PUT = 'PUT'
HTTP_PATCH = 'PATCH'
HTTP_DELETE = 'DELETE'
HTTP_OPTIONS = 'OPTIONS'

HTTP_METHODS = frozenset({
    HTTP_GET, HTTP_POST, HTTP_PUT, HTTP_PATCH, HTTP_DELETE, HTTP_OPTIONS
})

RE_HANDLER_METHOD = re.compile(r'^on_([a-z]+)$')
